@@ -13,7 +13,7 @@ import math
 version = csrbox.__version__
 logger = logging.getLogger(__name__)
 
-def csr_gen (isa_file, grp_file, custom_file, custom_attr_file, work_dir,
+def csr_gen (isa_file, grp_file, custom_file, debug_file, custom_attr_file, work_dir,
         probeinst, logging=False):
     '''
         Top level function which validates the sanity of the grouping yaml 
@@ -36,10 +36,30 @@ def csr_gen (isa_file, grp_file, custom_file, custom_attr_file, work_dir,
     
     isa_yaml = utils.load_yaml(isa_file)
     grp_yaml = utils.load_yaml(grp_file)
+    global debug_enabled
+    global compressed_enabled
+    global supervisor_enabled
+
+    if 'C' in isa_yaml['hart0']['ISA']:
+        compressed_enabled = True
+    else:
+        compressed_enabled = False
+
+    if 'S' in isa_yaml['hart0']['ISA']:
+        supervisor_enabled = True
+    else:
+        supervisor_enabled = False
 
     if custom_file :
         custom_yaml= utils.load_yaml(custom_file)
         isa_yaml['hart0'].update(custom_yaml['hart0'])
+    if debug_file:
+        debug_yaml = utils.load_yaml(debug_file)
+        isa_yaml['hart0'].update(debug_yaml['hart0'])
+        debug_enabled = True
+
+    else:
+        debug_enabled = False
 
     
     csrgrp_check.csrgrp_check(isa_yaml, grp_yaml, logging)
@@ -165,7 +185,7 @@ def groups_gen(isa_yaml, grp_yaml, attr_yaml, work_dir, probeinst, logging=False
         bsvfile.write(bsvstr)
         station_isa.clear()  
     physical_addr_sz = isa_yaml['hart0']['physical_addr_sz']
-    bsvstr = csr_top(csr_isa, grp_yaml, attr_yaml, rvxlen,
+    bsvstr = csr_top(csr_isa, grp_yaml, attr_yaml, rvxlen, 
              'csrbox', logging)
     filename = 'csrbox.bsv'
     bsvfile = open(work_dir + filename, 'w')
@@ -289,6 +309,14 @@ def csr_define(isa_yaml, work_dir, bsv, logging=False):
                  
                 definefile.write(address_temp.format(csr_name.upper(),csr_address))
 
+    if isa_yaml['hart0']['custom_exceptions'] is not None:
+        for ce in isa_yaml['hart0']['custom_exceptions']:
+            definefile.write('\n`define {0} {1}'.format(ce['cause_name'],
+                ce['cause_val']))
+    if isa_yaml['hart0']['custom_interrupts'] is not None:
+        for ce in isa_yaml['hart0']['custom_interrupts']:
+            definefile.write('\n`define {0} {1}'.format(ce['cause_name'],
+                ce['cause_val']))
     definefile.close()
 
 
@@ -490,18 +518,29 @@ def station_interface(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
         :type bsvstr: str
         :type work_dir: str
     '''
+    global debug_enabled
+    global compressed_enabled
+    global supervisor_enabled
     xlen = int(rvxlen[2:])
     bsvstr += interface_temp.format(pkg_name)
 
     if logging:
         logger.info('Initiating interface generation for : ' + pkg_name)
     # define additional methods for certain csrs
-    req_regs = []
+    req_ifcs = []
     for csrname in isa_yaml:
-        if csrname.upper() in additional_ifc:
-            if additional_ifc[csrname.upper()] not in req_regs:
-                bsvstr += textwrap.indent(additional_ifc[csrname.upper()],'    ')
-                req_regs.append(additional_ifc[csrname.upper()])
+        if csrname.lower() in attr_yaml and 'additional_method' in attr_yaml[csrname.lower()]:
+            for add_ifc in attr_yaml[csrname.lower()]['additional_method']:
+                if '::' in add_ifc:
+                    cond = add_ifc.split('::')[0]
+                    if cond:
+                        add_ifc = add_ifc.split('::')[1]
+                    else:
+                        add_ifc = ''
+                if add_ifc not in req_ifcs:
+                    bsvstr +=\
+                        textwrap.indent(add_ifc, '    ') + '\n'
+                    req_ifcs.append(add_ifc)
 
     # define a value method for each csr.
     for csrname in isa_yaml:
@@ -517,11 +556,18 @@ def station_interface(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
             if  logging:
                 logger.debug('-- Creating action methods for ' + csrname)
             size = attr_yaml[csrname]['size']
-            arguments = attr_yaml[csrname]['action_method']['def_arguments']
-            arguments = arguments.replace('xlen',str(xlen))
-            arguments = arguments.replace('size',str(size))
-            arguments = textwrap.indent(arguments, '    ')
-            bsvstr +=('\n    '+arguments)
+            for arguments in attr_yaml[csrname]['action_method']['def_arguments']:
+                arguments = arguments.replace('xlen',str(xlen))
+                arguments = arguments.replace('size',str(size))
+                if '::' in arguments:
+                    cond, args = re.findall(r'^(.*?)::(.*?)$',arguments)[0]
+                    if eval(cond):         
+                        args = textwrap.indent(args, '')
+                    else:
+                        args = ''
+                else:
+                    args = arguments
+                bsvstr +=('\n    '+args)
     bsvstr +=('\n')
 
     # create action methods of dependent csrs
@@ -611,7 +657,7 @@ def station_module(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
         if 'mod_attributes' in attr_yaml[csrname]:
             for modattr in attr_yaml[csrname]['mod_attributes']:
                 attr_str += '  ' + modattr + '\n'
-    bsvstr += (module_temp.format(pkg_name, attr_str))
+    bsvstr += (module_temp.format(pkg_name, attr_str, ''))
     if logging:
         logger.info('Module created for : ' + pkg_name)
 
@@ -680,13 +726,6 @@ def station_methods(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
     xlen = int(rvxlen[2:])
     if logging:
         logger.info('Creating method definitions for : ' + pkg_name)
-    # define additional methods for certain csrs
-    req_regs = []
-    for csrname in isa_yaml:
-        if csrname.upper() in additional_ifc_dec:
-            if additional_ifc_dec[csrname.upper()] not in req_regs:
-                bsvstr += textwrap.indent(additional_ifc_dec[csrname.upper()],'    ')
-                req_regs.append(additional_ifc_dec[csrname.upper()])
     
     # default value methods
     for csrname in isa_yaml:
@@ -694,7 +733,8 @@ def station_methods(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
             logger.debug('-- Defining value methods for : ' + csrname)
         if attr_yaml[csrname]['value_method']['required']:
             if 'dec_body' in attr_yaml[csrname]['value_method']:
-                bsvstr += methodv_def_temp.format(csrname, '', attr_yaml[csrname]['value_method']['dec_body'])
+                for dec in attr_yaml[csrname]['value_method']['dec_body']:
+                    bsvstr += methodv_def_temp.format(csrname, '', dec)
             else:
                 bsvstr += methodv_def_temp.format(csrname, 'rg_', csrname)
     
@@ -704,9 +744,16 @@ def station_methods(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
             logger.debug('-- Defining custom Action methods for : ' + csrname)
         if attr_yaml[csrname]['action_method']['required']:
             size = attr_yaml[csrname]['size']
-            _str = attr_yaml[csrname]['action_method']['dec_body']
-            _str = _str.replace('size', str(size))
-            bsvstr += textwrap.indent( _str+ '\n', '    ')
+            for _str in attr_yaml[csrname]['action_method']['dec_body']:
+                _str = _str.replace('size', str(size))
+                _str = _str.replace('xlen', str(xlen))
+                if '::' in _str:
+                    cond = _str.split('::')[0]
+                    if eval(cond):
+                        _str = _str.split('::')[1]
+                    else:
+                        _str = ''
+                bsvstr += textwrap.indent( _str+ '\n', '    ')
     
     # create action methods of dependent csrs
     incomming_reads = []
@@ -787,6 +834,7 @@ def station_req(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
         :type bsvstr: str
         :type work_dir: str
     '''
+    global compressed_enabled
 
     if logging:
         logger.info('Generating Core request routine for : ' + pkg_name)
@@ -824,7 +872,16 @@ def station_req(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
                 shadow_csr is None or shadow_csr in isa_yaml):
                 if logging:
                     logger.debug('-- Creating custom logic for : ' + csrname)
-                _t = attr_yaml[csrname]['csr_op'].replace('size',str(size))
+                for csrop in attr_yaml[csrname]['csr_op']:
+                    if '::' in csrop:
+                        cond = csrop.split('::')[0]
+                        if eval(cond):
+                            cust_csrop = csrop.split('::')[1]
+                        else:
+                            cust_csrop = ''
+                    else:
+                        cust_csrop = csrop
+                _t = cust_csrop.replace('size',str(size))
                 _t = _t.replace('xlen',str(xlen))
                 _t = textwrap.indent(_t, '            ') + '\n'
                 bsvstr += _t
@@ -882,10 +939,21 @@ def station_regs(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
     # creating additional required registers
     req_regs = []
     for csrname in isa_yaml:
-        if csrname.upper() in additional_regs:
-            if additional_regs[csrname.upper()] not in req_regs:
-                bsvstr += textwrap.indent(additional_regs[csrname.upper()],'    ')
-                req_regs.append(additional_regs[csrname.upper()])
+        if csrname in attr_yaml and 'additional_regs' in attr_yaml[csrname.lower()]:
+            for add_reg in attr_yaml[csrname.lower()]['additional_regs']:
+                if '::' in add_reg:
+                    cond = add_reg.split('::')[0]
+                    if eval(cond):
+                        add_reg = add_reg.split('::')[1]
+                    else:
+                        add_reg = ''
+                if add_reg not in req_regs:
+                    bsvstr += '\n'+textwrap.indent(add_reg,'    ')
+                    req_regs.append(add_reg)
+#        if csrname.upper() in additional_regs:
+#            if additional_regs[csrname.upper()] not in req_regs:
+#                bsvstr += textwrap.indent(additional_regs[csrname.upper()],'    ')
+#                req_regs.append(additional_regs[csrname.upper()])
 
     # create wires of dependent csrs
     incomming_reads = []
@@ -1045,6 +1113,8 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
         :type rvxlen: str
         :type pkgname: str
     '''
+    global debug_enabled
+    global compressed_enabled
 
     grp_list = sorted(list(grp_yaml))
     xlen = int(rvxlen[2:])
@@ -1076,6 +1146,14 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     mideleg_grp = find_group(grp_yaml, 'mideleg')
     sedeleg_grp = find_group(grp_yaml, 'sedeleg')
     sideleg_grp = find_group(grp_yaml, 'sideleg')
+    dpc_grp = find_group(grp_yaml,'dpc')
+    dcsr_grp_name = find_group(grp_yaml,'dcsr')
+    dtvec_grp = find_group(grp_yaml,'dtvec')
+
+    if debug_enabled:
+        if dpc_grp is None or dcsr_grp_name is None or dtvec_grp is None:
+            logger.error('Debug is enabled but one of dpc, dcsr and/or dtvec is not available in any group')
+            sys.exit(1)
     
     pmp_entries = 0
 
@@ -1114,14 +1192,22 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
 
 
     bsvstr += interface_temp.format(pkgname)
-    bsvstr += '''    interface Sbread sbread;'''
+    bsvstr += '''    interface Sbread sbread;\n'''
     # define additional required interfaces
-    req_regs = []
+    req_ifcs = []
     for csrname in isa_yaml:
-        if csrname.upper() in additional_ifc:
-            if additional_ifc[csrname.upper()] not in req_regs:
-                bsvstr += textwrap.indent(additional_ifc[csrname.upper()],'    ')
-                req_regs.append(additional_ifc[csrname.upper()])
+        if csrname.lower() in attr_yaml and 'additional_top_method' in attr_yaml[csrname.lower()]:
+            for add_ifc in attr_yaml[csrname.lower()]['additional_top_method']:
+                if '::' in add_ifc:
+                    cond = add_ifc.split('::')[0]
+                    if cond:
+                        add_ifc = add_ifc.split('::')[1]
+                    else:
+                        add_ifc = ''
+                if add_ifc not in req_ifcs:
+                    bsvstr +=\
+                        textwrap.indent(add_ifc, '    ') + '\n'
+                    req_ifcs.append(add_ifc)
         if 'pmpaddr' in csrname.lower():
             if isa_yaml[csrname]['rv'+str(xlen)]['accessible']:
                 pmp_entries += 1
@@ -1137,11 +1223,22 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
         if attr_yaml[csrname]['action_method']['required'] and \
                 attr_yaml[csrname]['action_method']['global']: 
             size = attr_yaml[csrname]['size']
-            arguments = attr_yaml[csrname]['action_method']['def_arguments']
-            arguments = arguments.replace('xlen',str(xlen))
-            arguments = arguments.replace('size',str(size))
-            arguments = textwrap.indent(arguments, '    ')
-            bsvstr +=('\n    '+arguments)
+            for arguments in attr_yaml[csrname]['action_method']['def_arguments']:
+                arguments = arguments.replace('xlen',str(xlen))
+                arguments = arguments.replace('size',str(size))
+                if '::' in arguments:
+                    cond, args = re.findall(r'^(.*?)::(.*?)$',arguments)[0]
+                    if eval(cond):         
+                        args = textwrap.indent(args, '')
+                    else:
+                        args = ''
+                else:
+                    args = arguments
+                bsvstr +=('\n    '+args)
+
+    # debug specific interfaces
+    if debug_enabled:
+        bsvstr += method_top_debug_dec;
     bsvstr += ('\n')
     bsvstr += (method_top_generic.format(xlen))
     bsvstr += (end_interface_temp)
@@ -1154,7 +1251,11 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     if 'minstret' in isa_yaml:
         if isa_yaml['minstret']['rv'+str(xlen)]['accessible']:
             mod_attr_str = '(*conflict_free="ma_incr_minstret,sbread.mv_csr_minstret"*)'
-    bsvstr += module_temp.format(pkgname,mod_attr_str) + '\n'
+    if debug_enabled:
+        extra_reset = '#(Reset debug_reset)'
+    else:
+        extra_reset = ''
+    bsvstr += module_temp.format(pkgname,mod_attr_str, extra_reset) + '\n'
 
     if logging:
         logger.debug('-- Instatiating groups')
@@ -1170,6 +1271,16 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     /*doc:reg: holds the current privilege level*/
     Reg#(Privilege_mode) rg_prv <- mkReg(Machine);
 '''.format(misa_grp)
+    
+    if debug_enabled:
+        bsvstr += '''
+    /*doc:reg:The register indicates that the core is halted*/
+    Reg#(Bit#(1)) rg_core_halted <- mkReg(0);
+
+    /*doc:reg: this register indicates if the core has been reset by the
+    debugger*/
+    Reg#(Bit#(1)) rg_core_has_reset <- mkRegA(0, reset_by debug_reset);
+'''
 
     anyhit = ''
     anydata = ''
@@ -1183,11 +1294,20 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     # additional wires/regs
     req_regs = []
     for csrname in isa_yaml:
-        if csrname.upper() in additional_regs:
-            if additional_regs[csrname.upper()] not in req_regs:
-                bsvstr +=\
-                textwrap.indent(additional_regs[csrname.upper()],'    ') + '\n'
-                req_regs.append(additional_regs[csrname.upper()])
+        if csrname.lower() in attr_yaml and 'additional_top_regs' in attr_yaml[csrname.lower()]:
+            grp_name = find_group(grp_yaml, csrname.upper())
+            for add_reg in attr_yaml[csrname.lower()]['additional_top_regs']:
+                add_reg = add_reg.replace('grp_name',str(grp_name))
+                if '::' in add_reg:
+                    cond = add_reg.split('::')[0]
+                    if cond:
+                        add_reg = add_reg.split('::')[1]
+                    else:
+                        add_reg = ''
+                if add_reg not in req_regs and add_reg != '' :
+                    bsvstr +=\
+                        textwrap.indent(add_reg, '    ') + '\n'
+                    req_regs.append(add_reg)
 
     # req=response connection across stations
     if logging:
@@ -1198,15 +1318,13 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     # -- connect rg_prv to all stations
     for grp in grp_yaml:
         bsvstr += temp_prv_connection.format(grp)
-    
-    req_ifc = []
-    for csrname in isa_yaml:
-        if csrname.upper() in additional_topifc_dec:
-            grp_name = find_group(grp_yaml, csrname.upper())
-            if (grp_name, additional_topifc_dec[csrname.upper()]) not in req_ifc:
-                bsvstr +='\n'+\
-                textwrap.indent(additional_topifc_dec[csrname.upper()].format(grp_name),'    ')
-                req_ifc.append((grp_name,additional_topifc_dec[csrname.upper()]))
+
+    if debug_enabled:
+        bsvstr += '''
+    rule rl_core_reset;
+        rg_core_has_reset <= 1;
+    endrule: rl_core_reset
+'''
     
     # create action methods of dependent csrs
     if logging:
@@ -1274,23 +1392,33 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
         bsvstr += method_top_trap_sedeleg.format(sedeleg_grp, sideleg_grp)
     if (usertraps or supervisor) and medeleg_grp is not None:
         bsvstr += method_top_trap_medeleg.format(medeleg_grp, mideleg_grp)
-    if usertraps:
-        bsvstr += method_top_trap_user_temp.format(utval_grp, uepc_grp, ucause_grp,
-            mstatus_grp, utvec_grp)
+    if debug_enabled:
+        bsvstr += method_top_trap_debug_temp.format(dpc_grp, dcsr_grp_name,
+                dtvec_grp)
     if supervisor:
         bsvstr += method_top_trap_supervisor_temp.format(stval_grp, sepc_grp, scause_grp,
             mstatus_grp, stvec_grp)
+    if usertraps:
+        bsvstr += method_top_trap_user_temp.format(utval_grp, uepc_grp, ucause_grp,
+            mstatus_grp, utvec_grp)
     bsvstr += method_top_trap_machine_temp.format(mtval_grp, mepc_grp, mcause_grp,
             mstatus_grp, mtvec_grp)
     
     # define additional required interfaces
     req_ifc = []
     for csrname in isa_yaml:
-        if csrname.upper() in additional_ifc_dec:
-            if additional_ifc_dec[csrname.upper()] not in req_ifc:
-                bsvstr +=\
-                textwrap.indent(additional_ifc_dec[csrname.upper()],'    ') + '\n'
-                req_ifc.append(additional_ifc_dec[csrname.upper()])
+        if csrname.lower() in attr_yaml and 'additional_top_method_body' in attr_yaml[csrname.lower()]:
+            for add_ifc in attr_yaml[csrname.lower()]['additional_top_method_body']:
+                if '::' in add_ifc:
+                    cond = add_ifc.split('::')[0]
+                    if cond:
+                        add_ifc = add_ifc.split('::')[1]
+                    else:
+                        add_ifc = ''
+                if add_ifc not in req_ifc:
+                    bsvstr +=\
+                        textwrap.indent(add_ifc, '    ') + '\n'
+                    req_ifc.append(add_ifc)
 
     # pmp method definitions
     pmpaddr_str = ''
@@ -1340,11 +1468,25 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
         logger.debug('Top CSR Interface: ' + csrname)
         if attr_yaml[csrname]['action_method']['required'] and \
                 attr_yaml[csrname]['action_method']['global']: 
-            defs =  attr_yaml[csrname]['action_method']['def_arguments']
+#            defs =  attr_yaml[csrname]['action_method']['def_arguments']
             csr_grp = find_group(grp_yaml, csrname.upper())
-            methods = re.findall(r'\s*method\s*Action\s*(.*?)[\s(]', defs)
-            for x in methods:
-                bsvstr += methoda_top_temp.format(x, csr_grp)
+            for arguments in attr_yaml[csrname]['action_method']['def_arguments']:
+                arguments = arguments.replace('xlen',str(xlen))
+                arguments = arguments.replace('size',str(size))
+                if '::' in arguments:
+                    cond, args = re.findall(r'^(.*?)::(.*?)$',arguments)[0]
+                    if eval(cond):         
+                        args = textwrap.indent(args, '')
+                    else:
+                        args = ''
+                else:
+                    args = arguments
+                methods = re.findall(r'\s*method\s*Action\s*(.*?)[\s(]', args)
+                for x in methods:
+                    bsvstr += methoda_top_temp.format(x, csr_grp)
+    # debug specific interfaces
+    if debug_enabled:
+        bsvstr += method_top_debug_def;
             
 
     # -- end the module
